import sys

import pika
import pika.credentials
import pika.exceptions


class Connector:
    DIRECT_TYPE = 'direct'
    FANOUT_TYPE  = 'fanout'
    TOPIC_TYPE = 'topic'
    HEADERS_TYPE = 'headers'

    DELAY_TYPE = 'x-delayed-message'

    def __init__(self,
                 connection_parameters: pika.ConnectionParameters,
                 exchange: str,
                 queue: str,
                 routing_key: str = None,
                 prefetch_count: int = 1,
                 exchange_type: str = DIRECT_TYPE,
                 use_delay: bool = False):
        self.connection_parameters = connection_parameters
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.queue = queue
        self.routing_key = routing_key or queue
        self.prefetch_count = prefetch_count
        self.connection = None
        self.channel = None
        self.declared_queue = None
        self.use_delay = use_delay

    def create_connection(self, declare=True):
        self.connection = pika.BlockingConnection(self.connection_parameters)
        self.channel = self.connection.channel()
        if declare:
            self.declared_queue = self.channel.queue_declare(queue=self.queue, durable=True)
            if self.use_delay:
                exchange_type = self.DELAY_TYPE
                arguments = {"x-delayed-type": self.exchange_type}
            else:
                exchange_type = self.exchange_type
                arguments = None
            self.channel.exchange_declare(
                exchange=self.exchange, exchange_type=exchange_type, durable=True, arguments=arguments
            )
            self.channel.queue_bind(self.queue, self.exchange, routing_key=self.routing_key)
            if self.prefetch_count:
                self.channel.basic_qos(prefetch_count=self.prefetch_count)

    def close(self):
        if self.connection:
            self.connection.close()


class Reconnector(Connector):
    def __init__(self, connection_parameters: pika.ConnectionParameters, *args, **kwargs):
        connection_parameters.connection_attempts = sys.maxsize
        super().__init__(connection_parameters, *args, **kwargs)



import datetime
import decimal
import json


class DateTimeDecimalJSONEncoder(json.encoder.JSONEncoder):

    class FakeFloat(float):
        def __init__(self, value):
            super().__init__()
            self._value = value

        def __repr__(self):
            return str(self._value)

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, decimal.Decimal):
            return self.FakeFloat(obj)
        else:
            return super(DateTimeDecimalJSONEncoder, self).default(obj)



import logging

import pika
import pika.exceptions

from .connectors import Connector, Reconnector
from . import dumpers


class Publisher:
    def __init__(self, connector: Connector):
        self.connector = connector
        self.connector.create_connection()

    def publish(self, **params):
        self.connector.channel.basic_publish(**params)

    def send_message(self, message, content_type='text/plain', delay=0, properties=None):
        dumper = dumpers.get_dumper(content_type)
        if dumper is None:
            raise RuntimeError("Can't dumper by content type %s" % content_type)
        message = dumper(message)
        properties = properties or {}
        properties.update({
            'delivery_mode': 2,
            'content_type': content_type
        })
        if delay != 0:
            assert self.connector.use_delay, u'Publisher must have delay support'
            properties.update({
                'headers': {
                    'x-delay': delay
                }
            })
        params = {
            "exchange": self.connector.exchange,
            "routing_key": self.connector.routing_key,
            "body": message,
            "properties": pika.BasicProperties(**properties)
        }
        self.publish(**params)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connector.close()


class ReconPublisher(Publisher):
    def __init__(self, connector: Reconnector, logger: logging.Logger = None):
        assert isinstance(connector, Reconnector), 'connector must be Reconnector instance'
        self.logger = logger
        super().__init__(connector)

    def publish(self, **params):
        while True:
            try:
                super(ReconPublisher, self).publish(**params)
                break
            except pika.exceptions.AMQPError:
                if self.logger:
                    self.logger.exception('RabbitMQ publisher exception')
                self.connector.create_connection()


class JSONPublisher(Publisher):
    def send_message(self, message, delay=0, properties=None):
        return super().send_message(
            message, content_type="application/json", delay=delay, properties=properties,
        )



import json
import pickle

from .encoders import DateTimeDecimalJSONEncoder


__dumpers_dict = {}


def register(content_type):
    def wrap_decor(dump_function):
        __dumpers_dict[content_type] = dump_function
        return dump_function
    return wrap_decor


@register('text/plain')
def pickle_dumper(data):
    return pickle.dumps(data)


@register('application/json')
def json_dumper(data):
    return json.dumps(data, cls=DateTimeDecimalJSONEncoder)


def get_dumper(content_type):
    return __dumpers_dict.get(content_type)
